#include <catch2/catch.hpp>

#include "Hardware/Adversaries.hpp"
#include "Hardware/Prepare.hpp"

namespace Hardware {
TEST_CASE("Test full verification with excluded probes (dom_indep_d1_exclude)", "[dom_indep_d1_exclude]") {
  const std::string library_file_name = "library.json";
  const std::string library_name = "NANG45";
  const std::string design_file_name = "ut/full/dom_indep_d1_exclude/dom_indep_d1_exclude.v";
  const std::string config_file_name =
      "ut/full/dom_indep_d1_exclude/dom_indep_d1_exclude.json";
  const std::string topmodule_name = "dom_indep_d1_exclude";
  const std::string result_folder_name = "ut/full/dom_indep_d1_exclude";

  Settings settings(config_file_name);

  SECTION("Robust Probing Model") {
    settings.relaxed_model = false;
    settings.order_of_test = GENERATE(1, 2);

    CircuitStruct circuit;
    Library library(library_file_name, library_name, settings.relaxed_model);
    Hardware::Read::DesignFile(design_file_name, topmodule_name, settings,
                              library, &circuit, 0, 0, 0, NULL);
    Hardware::Prepare::MakeCircuitDepth(library, &circuit);

    std::vector<SharedData> shared_data(settings.number_of_threads,
                                        SharedData(circuit, settings));
    Simulation simulation(circuit, settings);
    simulation.topmodule_name_ = topmodule_name;
    simulation.result_folder_name_ = result_folder_name;

    Hardware::Adversaries<RobustProbe> adversary(library, circuit, settings,
                                                 simulation);
    double maximum_leakage = adversary.EvaluateRobustProbingSecurity(
        library, circuit, settings, shared_data, simulation);

    if (settings.order_of_test == 1){
      // We expect no leakage in the first order.
      REQUIRE(maximum_leakage < 5.0);
    } else if (settings.order_of_test == 2){
      // But we expect leakage in the second order.
      REQUIRE(maximum_leakage >= 5.0);
    }
  }

  SECTION("Robust but Relaxed Probing Model") {
    settings.relaxed_model = true;
    settings.order_of_test = GENERATE(1, 2);

    CircuitStruct circuit;
    Library library(library_file_name, library_name, settings.relaxed_model);
    Hardware::Read::DesignFile(design_file_name, topmodule_name, settings,
                              library, &circuit, 0, 0, 0, NULL);
    Hardware::Prepare::MakeCircuitDepth(library, &circuit);

    std::vector<SharedData> shared_data(settings.number_of_threads,
                                        SharedData(circuit, settings));
    Simulation simulation(circuit, settings);
    simulation.topmodule_name_ = topmodule_name;
    simulation.result_folder_name_ = result_folder_name;

    Hardware::Adversaries<RelaxedProbe> adversary(library, circuit, settings,
                                                 simulation);
    double maximum_leakage = adversary.EvaluateRobustProbingSecurity(
        library, circuit, settings, shared_data, simulation);

    if (settings.order_of_test == 1){
      // We expect no leakage in the first order.
      REQUIRE(maximum_leakage < 5.0);
    } else if (settings.order_of_test == 2){
      // But we expect leakage in the second order.
      REQUIRE(maximum_leakage >= 5.0);
    }
  }
}
}  // namespace Hardware