module WDDL_AND (a, b, c);
  input [1:0] a;
  input [1:0] b;
  output [1:0] c;

  AND2_X1 U1 (.A1(a[1]), .A2(b[1]), .ZN(c[1]));
  OR2_X1 U2 (.A1(a[0]), .A2(b[0]), .ZN(c[0]));
endmodule

module WDDL_XOR (a, b, c);
  input [1:0] a;
  input [1:0] b;
  output [1:0] c;

  wire   [3:0] w;

  AND2_X1 U1 (.A1(a[1]), .A2(b[0]), .ZN(w[0]));
  AND2_X1 U2 (.A1(a[0]), .A2(b[1]), .ZN(w[1]));
  OR2_X1 U3 (.A1(w[0]), .A2(w[1]), .ZN(c[1]));
  OR2_X1 U4 (.A1(a[1]), .A2(b[0]), .ZN(w[2]));
  OR2_X1 U5 (.A1(a[0]), .A2(b[1]), .ZN(w[3]));
  AND2_X1 U6 (.A1(w[2]), .A2(w[3]), .ZN(c[0]));
endmodule

module sesym_indep_d1 (clk, a0, a0_bar, b0, b0_bar, z0, a1, a1_bar, b1, b1_bar, z1, r, r_bar);
  input clk;
  input r, r_bar;
  input a0, a0_bar; 
  input a1, a1_bar; 
  input b0, b0_bar; 
  input b1, b1_bar; 
  output [1:0] z0;
  output [1:0] z1;

  wire [1:0] c0, c1;
  wire u0, u1, u2, u3, v0, v1;
  wire u0_bar, u1_bar, u2_bar, u3_bar, v0_bar, v1_bar;

  WDDL_AND U0 ( .a({a0,a0_bar}), .b({b0,b0_bar}), .c({u0,u0_bar}) );
  WDDL_AND U1 ( .a({a0,a0_bar}), .b({b1,b1_bar}), .c({u1,u1_bar}) );
  WDDL_AND U2 ( .a({a1,a1_bar}), .b({b0,b0_bar}), .c({u2,u2_bar}) );
  WDDL_AND U3 ( .a({a1,a1_bar}), .b({b1,b1_bar}), .c({u3,u3_bar}) );

  WDDL_XOR U4 ( .a({u1,u1_bar}), .b({r,r_bar}), .c({v0,v0_bar}) );
  WDDL_XOR U5 ( .a({u2,u2_bar}), .b({r,r_bar}), .c({v1,v1_bar}) );
  WDDL_XOR U6 ( .a({u0,u0_bar}), .b({v0,v0_bar}), .c(c0) );
  WDDL_XOR U7 ( .a({u3,u3_bar}), .b({v1,v1_bar}), .c(c1) );

  DFF_X1 c0_0_reg ( .D(c0[0]), .CK(clk), .Q(z0[0]), .QN() );
  DFF_X1 c0_1_reg ( .D(c0[1]), .CK(clk), .Q(z0[1]), .QN() );
  DFF_X1 c1_0_reg ( .D(c1[0]), .CK(clk), .Q(z1[0]), .QN() );
  DFF_X1 c1_1_reg ( .D(c1[1]), .CK(clk), .Q(z1[1]), .QN() );
endmodule
