module WDDL_AND (a, b, c);
  input [1:0] a;
  input [1:0] b;
  output [1:0] c;

  AND2_X1 U1 (.A1(a[1]), .A2(b[1]), .ZN(c[1]));
  OR2_X1 U2 (.A1(a[0]), .A2(b[0]), .ZN(c[0]));
endmodule

module WDDL_XOR (a, b, c);
  input [1:0] a;
  input [1:0] b;
  output [1:0] c;

  wire   [3:0] w;

  AND2_X1 U1 (.A1(a[1]), .A2(b[0]), .ZN(w[0]));
  AND2_X1 U2 (.A1(a[0]), .A2(b[1]), .ZN(w[1]));
  OR2_X1 U3 (.A1(w[0]), .A2(w[1]), .ZN(c[1]));
  OR2_X1 U4 (.A1(a[1]), .A2(b[0]), .ZN(w[2]));
  OR2_X1 U5 (.A1(a[0]), .A2(b[1]), .ZN(w[3]));
  AND2_X1 U6 (.A1(w[2]), .A2(w[3]), .ZN(c[0]));
endmodule

module sesym_indep_d2 (clk, a0, a0_bar, b0, b0_bar, z0, a1, a1_bar, b1, b1_bar, z1, a2, a2_bar, b2, b2_bar, z2, r0, r0_bar, r1, r1_bar, r2, r2_bar);
  input clk;
  input r0, r0_bar;
  input r1, r1_bar;
  input r2, r2_bar;
  input a0, a0_bar; 
  input a1, a1_bar; 
  input a2, a2_bar; 
  input b0, b0_bar; 
  input b1, b1_bar; 
  input b2, b2_bar; 
  output [1:0] z0;
  output [1:0] z1;
  output [1:0] z2;

  wire i0, i0_bar, i1, i1_bar, i2, i2_bar, i3, i3_bar, i4, i4_bar, i5, i5_bar, i6, i6_bar, i7, i7_bar, i8, i8_bar;
  wire j0, j0_bar, j1, j1_bar, j2, j2_bar, j3, j3_bar, j4, j4_bar, j5, j5_bar;
  wire p0, p0_bar, p1, p1_bar, p2, p2_bar;
  wire [1:0] c0, c1, c2;

  WDDL_AND U0 ( .a({a0,a0_bar}), .b({b0,b0_bar}), .c({i0,i0_bar}) );
  WDDL_AND U1 ( .a({a0,a0_bar}), .b({b1,b1_bar}), .c({i1,i1_bar}) );
  WDDL_AND U2 ( .a({a0,a0_bar}), .b({b2,b2_bar}), .c({i2,i2_bar}) );
  WDDL_AND U3 ( .a({a1,a1_bar}), .b({b0,b0_bar}), .c({i3,i3_bar}) );
  WDDL_AND U4 ( .a({a1,a1_bar}), .b({b1,b1_bar}), .c({i4,i4_bar}) );
  WDDL_AND U5 ( .a({a1,a1_bar}), .b({b2,b2_bar}), .c({i5,i5_bar}) );
  WDDL_AND U6 ( .a({a2,a2_bar}), .b({b0,b0_bar}), .c({i6,i6_bar}) );
  WDDL_AND U7 ( .a({a2,a2_bar}), .b({b1,b1_bar}), .c({i7,i7_bar}) );
  WDDL_AND U8 ( .a({a2,a2_bar}), .b({b2,b2_bar}), .c({i8,i8_bar}) );

  WDDL_XOR U9  ( .a({i1,i1_bar}), .b({r0,r0_bar}), .c({j0,j0_bar}) );
  WDDL_XOR U10 ( .a({i2,i2_bar}), .b({r1,r1_bar}), .c({j1,j1_bar}) );
  WDDL_XOR U11 ( .a({i3,i3_bar}), .b({r0,r0_bar}), .c({j2,j2_bar}) );
  WDDL_XOR U12 ( .a({i5,i5_bar}), .b({r2,r2_bar}), .c({j3,j3_bar}) );
  WDDL_XOR U13 ( .a({i6,i6_bar}), .b({r1,r1_bar}), .c({j4,j4_bar}) );
  WDDL_XOR U14 ( .a({i7,i7_bar}), .b({r2,r2_bar}), .c({j5,j5_bar}) );

  WDDL_XOR U15 ( .a({j0,j0_bar}), .b({j1,j1_bar}), .c({p0,p0_bar}) );
  WDDL_XOR U16 ( .a({j2,j2_bar}), .b({j3,j3_bar}), .c({p1,p1_bar}) );
  WDDL_XOR U17 ( .a({j4,j4_bar}), .b({j5,j5_bar}), .c({p2,p2_bar}) );

  WDDL_XOR U18 ( .a({i0,i0_bar}), .b({p0,p0_bar}), .c(c0) );
  WDDL_XOR U19 ( .a({i4,i4_bar}), .b({p1,p1_bar}), .c(c1) );
  WDDL_XOR U20 ( .a({i8,i8_bar}), .b({p2,p2_bar}), .c(c2) );

  DFF_X1 U21 ( .D(c0[0]), .CK(clk), .Q(z0[0]), .QN() );
  DFF_X1 U22 ( .D(c0[1]), .CK(clk), .Q(z0[1]), .QN() );
  DFF_X1 U23 ( .D(c1[0]), .CK(clk), .Q(z1[0]), .QN() );
  DFF_X1 U24 ( .D(c1[1]), .CK(clk), .Q(z1[1]), .QN() );
  DFF_X1 U25 ( .D(c2[0]), .CK(clk), .Q(z2[0]), .QN() );
  DFF_X1 U26 ( .D(c2[1]), .CK(clk), .Q(z2[1]), .QN() );
endmodule
